/************************************************************************
 *
 * Basic Interval Arithmetic Subroutines Level 0
 * ---------------------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: Bias0.h,v 1.1 1997/09/23 11:50:35 knueppel Exp $
 *
 ************************************************************************/

#ifndef __BIAS0__
#define __BIAS0__

#include <BIAS/BiasType.h>

#define __BIASINFSUP__
#define __BIASIEEE__

/************************************************************************
 *  Definition of Constants, Types, and Variables
 ************************************************************************/

extern REAL     BiasEpsilon;            /* min {x >= 0 : 1 + x > 1}     */
extern REAL     BiasEta;                /* min {x > 0}                  */
extern REAL     BiasNaN;                /* NaN                          */
extern REAL     BiasPosInf;             /* +oo                          */
extern REAL     BiasNegInf;             /* -oo                          */

/************************************************************************
 *  Function Prototypes
 ************************************************************************/
#if defined (__cplusplus)
extern "C" {
#endif

VOID    BiasRoundUp     (VOID);
VOID    BiasRoundDown   (VOID);
VOID    BiasRoundNear   (VOID);

VOID    BiasInit        (VOID);

REAL    BiasPredR       (const REAL         * const pa);
INT     BiasPredI       (      BIASINTERVAL * const pR,
                         const BIASINTERVAL * const pA);

REAL    BiasSuccR       (const REAL         * const pa);
VOID    BiasSuccI       (BIASINTERVAL       * const pR,
                         const BIASINTERVAL * const pA);

VOID    BiasAddRR       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasAddRI       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasAddII       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasSubRR       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasSubRI       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasSubIR       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const REAL         * const pb);
VOID    BiasSubII       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasMulRR       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasMulRI       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasMulII       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasDivRR       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasDivRI       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasDivIR       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const REAL         * const pb);
VOID    BiasDivII       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasMacRR       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasMacRI       (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasMacII       (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasNeg         (BIASINTERVAL       * const pR,
                         const BIASINTERVAL * const pA);

REAL    BiasMid         (const BIASINTERVAL * const pA);
VOID    BiasMidRad      (REAL               * const pm,
                         REAL               * const pr,
			 const BIASINTERVAL * const pA);
REAL    BiasDiam        (const BIASINTERVAL * const pA);

REAL    BiasAbs         (const BIASINTERVAL * const pA);

REAL    BiasDistRI      (const REAL         * const pa,
                         const BIASINTERVAL * const pB);
REAL    BiasDistII      (const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

INT     BiasIntersection(BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

VOID    BiasHullR       (BIASINTERVAL       * const pR,
                         const REAL         * const pa);
VOID    BiasHullRR      (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const REAL         * const pb);
VOID    BiasHullRI      (BIASINTERVAL       * const pR,
			 const REAL         * const pa,
                         const BIASINTERVAL * const pB);
VOID    BiasHullII      (BIASINTERVAL       * const pR,
			 const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

INT     BiasInR         (const REAL         * const pa,
                         const BIASINTERVAL * const pB);
INT     BiasInI         (const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

INT     BiasInInteriorR (const REAL         * const pa,
                         const BIASINTERVAL * const pB);
INT     BiasInInteriorI (const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

INT     BiasIsEqual     (const BIASINTERVAL * const pA,
                         const BIASINTERVAL * const pB);

#define BiasInf(pA)     ((pA)->inf)
#define BiasSup(pA)     ((pA)->sup)

#define BiasAddIR(pR, pA, pb)    BiasAddRI (pR, pb, pA)
#define BiasMulIR(pR, pA, pb)    BiasMulRI (pR, pb, pA)
#define BiasMacIR(pR, pA, pb)    BiasMacRI (pR, pb, pA)
#define BiasDistIR(pA, pb)       BiasDistRI (pb, pA)
#define BiasHullIR(pR, pA, pb)   BiasHullRI (pR, pb, pA)

#if defined (__cplusplus)
}
#endif

#endif /* __BIAS0__ */
