/************************************************************************
 *
 * Basic Interval Arithmetic Subroutines Test Program
 * --------------------------------------------------
 *
 * Copyright (c) 1997 Olaf Knueppel
 *
 * $Id: BiasTest.c,v 1.3 1998/10/07 11:06:43 knueppel Exp $
 *
 ************************************************************************/

#include <BIAS/Bias0.h>
#include <stdio.h>
#include <stdlib.h>

static void error (char *s)
{
  fprintf (stderr, "ERROR: %s\n", s);
}

static void Dump (char *s, double Result)
{
  int i;
  unsigned char *p = (unsigned char *) & Result;

  fprintf (stderr, "%+2.2f %-8.8s: ", Result, s);
  for (i = 0; i < sizeof (double); i++)
    fprintf (stderr, "%02X ", *p++);
  fprintf (stderr, "\n");
}

static void DumpValues (void)
{
  static volatile double One = 1.0;
  static volatile double Ten = 10.0;

  volatile double Result;

  BiasRoundNear ();
  Result = One / Ten;
  Dump ("Nearest", Result);
  BiasRoundUp ();
  Result = One / Ten;
  Dump ("Up", Result);
  BiasRoundDown ();
  Result = One / Ten;
  Dump ("Down", Result);

  BiasRoundNear ();
  Result = -One / Ten;
  Dump ("Nearest", Result);
  BiasRoundUp ();
  Result = -One / Ten;
  Dump ("Up", Result);
  BiasRoundDown ();
  Result = -One / Ten;
  Dump ("Down", Result);
}

int main (int argc, char *argv[])
{
  int rc;
  BIASINTERVAL x, y;
  static double One      =  1.0;
  static double MinusOne = -1.0;
  static double Three    =  3.0;
  static double OneThird;

  BiasInit ();
  OneThird = One / Three;

  BiasHullRR (& x, & MinusOne, & One);
  BiasDivIR (& y, & x, & Three);

  rc = 0;

  if (BiasInf (& y) >= -OneThird) {
    error ("illegal downward rounding");
    rc = 1;
  }
  if (BiasSup (& y) <=  OneThird) {
    error ("illegal upward rounding");
    rc = 1;
  }

  if (argc > 1) DumpValues ();

  return rc;
}
