/************************************************************************
 *
 * Definition of type COMPLEX
 * --------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: Complex.h,v 1.1 1997/09/23 11:50:52 knueppel Exp $
 *
 ************************************************************************/

#ifndef __COMPLEX__
#define __COMPLEX__

#include <Configuration.h>
#include <iostream.h>

class COMPLEX {
  REAL re;
  REAL im;
public:
  COMPLEX () { }
  COMPLEX (CONST REAL r) { re = r; im = 0.0; }
  COMPLEX (CONST REAL r, CONST REAL i) { re = r; im = i; }
  COMPLEX & operator = (CONST REAL x) { re = x; im = 0.0; return *this; }
  COMPLEX & operator += (CONST REAL x) { re += x; return *this; }
  COMPLEX & operator -= (CONST REAL x) { re -= x; return *this; }
  COMPLEX & operator *= (CONST REAL x) { re *= x; im *= x; return *this; }
  COMPLEX & operator /= (CONST REAL x) { re /= x; im /= x; return *this; }
  COMPLEX & operator = (CONST COMPLEX & x)
    { re = Re(x); im = Im(x); return *this; }
  COMPLEX & operator += (CONST COMPLEX & x)
    { re += Re(x); im += Im(x); return *this; }
  COMPLEX & operator -= (CONST COMPLEX & x)
    { re -= Re(x); im -= Im(x); return *this; }
  COMPLEX & operator *= (CONST COMPLEX & x)
    { REAL t = re;
      re = re * Re(x) - im * Im(x);
      im = im * Re(x) + t  * Im(x);
      return *this;
    }
  COMPLEX & operator /= (CONST COMPLEX & x);
  friend COMPLEX operator + (CONST COMPLEX & x) { return (COMPLEX &) x; }
  friend COMPLEX operator - (CONST COMPLEX & x)
    { return COMPLEX (-Re(x), -Im(x)); }
  friend REAL Re (CONST COMPLEX & x) { return x.re; }
  friend REAL Im (CONST COMPLEX & x) { return x.im; }
  friend COMPLEX operator + (CONST COMPLEX & x, CONST REAL y)
    { return COMPLEX (Re(x) + y, Im(x)); }
  friend COMPLEX operator + (CONST REAL x, CONST COMPLEX & y)
    { return COMPLEX (x + Re(y), Im(y)); }
  friend COMPLEX operator + (CONST COMPLEX & x, CONST COMPLEX & y)
    { return COMPLEX (Re(x) + Re(y), Im(x) + Im(y)); }
  friend COMPLEX operator - (CONST COMPLEX & x, CONST REAL y)
    { return COMPLEX (Re(x) - y, Im(x)); }
  friend COMPLEX operator - (CONST REAL x, CONST COMPLEX & y)
    { return COMPLEX (x - Re(y), -Im(y)); }
  friend COMPLEX operator - (CONST COMPLEX & x, CONST COMPLEX & y)
    { return COMPLEX (Re(x) - Re(y), Im(x) - Im(y)); }
  friend COMPLEX operator * (CONST COMPLEX & x, CONST REAL y)
    { return COMPLEX (Re(x) * y, Im(x) * y); }
  friend COMPLEX operator * (CONST REAL x, CONST COMPLEX & y)
    { return COMPLEX (x * Re(y), x * Im(y)); }
  friend COMPLEX operator * (CONST COMPLEX & x, CONST COMPLEX & y)
    { return COMPLEX (Re(x) * Re(y) - Im(x) * Im(y),
		      Im(x) * Re(y) + Im(y) * Re(x)); }
  friend COMPLEX operator / (CONST COMPLEX & x, CONST REAL y)
    { return COMPLEX (Re(x) / y, Im(x) / y); }
  friend COMPLEX operator / (CONST REAL x, CONST COMPLEX & y);
  friend COMPLEX operator / (CONST COMPLEX & x, CONST COMPLEX & y);
  friend INT operator == (CONST COMPLEX & x, CONST REAL y)
    { return (Re(x) == y) && (Im(x) == 0.0); }
  friend INT operator == (CONST REAL x, CONST COMPLEX & y)
    { return (x == Re(y)) && (Im(y) == 0.0); }
  friend INT operator == (CONST COMPLEX & x, CONST COMPLEX & y)
    { return (Re(x) == Re(y)) && (Im(x) == Im(y)); }
  friend INT operator != (CONST COMPLEX & x, CONST REAL y)
    { return (Re(x) != y) || (Im(x) != 0.0); }
  friend INT operator != (CONST REAL x, CONST COMPLEX & y)
    { return (x != Re(y)) || (Im(y) != 0.0); }
  friend INT operator != (CONST COMPLEX & x, CONST COMPLEX & y)
    { return (Re(x) != Re(y)) || (Im(x) != Im(y)); }
  friend COMPLEX Conjg (CONST COMPLEX & x)
    { return COMPLEX (Re(x), -Im(x)); }
  friend REAL Abs (CONST COMPLEX & x);
  friend COMPLEX Sqrt (CONST COMPLEX & x);
  friend ostream & operator << (ostream & o, CONST COMPLEX & x);
  friend istream & operator >> (istream & i, COMPLEX & x);
};

extern COMPLEX I;

#endif /* __COMPLEX__ */
