/************************************************************************
 *
 * Definition of standard functions (REAL and INTERVAL)
 * ----------------------------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: Functions.h,v 1.1 1997/09/23 11:50:53 knueppel Exp $
 *
 ************************************************************************/

#ifndef __FUNCTIONS__
#define __FUNCTIONS__

#include <Interval.h>
#include <math.h>

class Constant {
public:
  static REAL Pi;
  static REAL TwoPi;
  static REAL PiHalf;
  static REAL PiQuarter;
  static REAL e;
  static REAL Sqrt2;
  static REAL InvSqrt2;
  static REAL Ln10;
  Constant ();
  ~Constant () { }
};

inline REAL Sin    (REAL x) { return sin(x); }
inline REAL Cos    (REAL x) { return cos(x); }
inline REAL Tan    (REAL x) { return tan(x); }
inline REAL Cot    (REAL x) { return -tan(x + Constant::PiHalf); }
inline REAL ArcSin (REAL x) { return asin(x); }
inline REAL ArcCos (REAL x) { return acos(x); }
inline REAL ArcTan (REAL x) { return atan(x); }
inline REAL ArcCot (REAL x) { return atan(-x) + Constant::PiHalf; }

inline REAL Sinh   (REAL x) { return sinh(x); }
inline REAL Cosh   (REAL x) { return cosh(x); }
inline REAL Tanh   (REAL x) { return tanh(x); }
inline REAL Coth   (REAL x) { return 1.0 / tanh(x); }
       REAL ArSinh (REAL x);
       REAL ArCosh (REAL x);
       REAL ArTanh (REAL x);
       REAL ArCoth (REAL x);

inline REAL Exp    (REAL x) { return exp(x); }
inline REAL Log    (REAL x) { return log(x); }
inline REAL Log10  (REAL x) { return log10(x); }

inline REAL Abs    (REAL x) { return fabs(x); }
inline REAL Sqr    (REAL x) { return x*x; }
inline REAL Sqrt   (REAL x) { return sqrt(x); }
inline REAL Root   (REAL x, INT n) { return pow(x, 1.0 / (REAL) n); }

       REAL Power  (REAL x, INT n);
inline REAL Power  (REAL x, REAL y) { return pow(x, y); }

inline REAL Min    (REAL x, REAL y) { return (x < y) ? x : y; }
inline REAL Max    (REAL x, REAL y) { return (x > y) ? x : y; }

       INTERVAL Sin    (CONST INTERVAL & x);
       INTERVAL Cos    (CONST INTERVAL & x);
       INTERVAL Tan    (CONST INTERVAL & x);
       INTERVAL Cot    (CONST INTERVAL & x);
       INTERVAL ArcSin (CONST INTERVAL & x);
       INTERVAL ArcCos (CONST INTERVAL & x);
       INTERVAL ArcTan (CONST INTERVAL & x);
       INTERVAL ArcCot (CONST INTERVAL & x);

       INTERVAL Sinh   (CONST INTERVAL & x);
       INTERVAL Cosh   (CONST INTERVAL & x);
       INTERVAL Tanh   (CONST INTERVAL & x);
       INTERVAL Coth   (CONST INTERVAL & x);
       INTERVAL ArSinh (CONST INTERVAL & x);
       INTERVAL ArCosh (CONST INTERVAL & x);
       INTERVAL ArTanh (CONST INTERVAL & x);
       INTERVAL ArCoth (CONST INTERVAL & x);

       INTERVAL Exp    (CONST INTERVAL & x);
       INTERVAL Log    (CONST INTERVAL & x);
       INTERVAL Log10  (CONST INTERVAL & x);

       INTERVAL IAbs   (CONST INTERVAL & x);
       INTERVAL Sqr    (CONST INTERVAL & x);
       INTERVAL Sqrt   (CONST INTERVAL & x);
       INTERVAL Root   (CONST INTERVAL & x, INT n);

       INTERVAL Power  (CONST INTERVAL & x, INT n);
       INTERVAL Power  (CONST INTERVAL & x, CONST INTERVAL & y);

#endif /* __FUNCTIONS__ */
