/************************************************************************
 *
 * Definition of type INTEGER_MATRIX
 * ---------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: IntegerMatrix.h,v 1.2 1999/07/28 08:55:46 knueppel Exp $
 *
 ************************************************************************/

#ifndef __INTEGER_MATRIX__
#define __INTEGER_MATRIX__

#include <Configuration.h>
#include <Error.h>
#include <IntegerVector.h>

class INTEGER_MATRIX {
  INT nRows;
  INT nCols;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  INT *theElements;
  INTEGER_MATRIX ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nRows = nCols = 0; theElements = NULL; }
  INTEGER_MATRIX (INT r, INT c)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckMatrixRange (r, c);
      nRows = r; nCols = c;
      theElements = new INT[r * c];
      CheckMemory (theElements);
    }
  ~INTEGER_MATRIX ()
    { if (theElements != NULL) delete theElements; }
  INTEGER_MATRIX (CONST INTEGER_MATRIX &);
  INT & operator () (INT i, INT j) CONST
    { CheckMatrixIndex (i, j, *this); return theElements[(i-1) * nCols + (j-1)]; }
  INTEGER_MATRIX & operator =  (CONST INTEGER_MATRIX &);
  INTEGER_MATRIX & operator += (CONST INTEGER_MATRIX &);
  INTEGER_MATRIX & operator -= (CONST INTEGER_MATRIX &);
  INTEGER_MATRIX & operator *= (INT);
  INTEGER_MATRIX & operator /= (INT);
  friend INT RowDimension (CONST INTEGER_MATRIX & x) { return x.nRows; }
  friend INT ColDimension (CONST INTEGER_MATRIX & x) { return x.nCols; }
  friend VOID Resize (INTEGER_MATRIX &, INT, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (INTEGER_MATRIX & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (INTEGER_MATRIX & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (INTEGER_MATRIX & x) { }
  friend VOID MakePermanent (INTEGER_MATRIX & x) { }
#endif
  friend VOID Clear (INTEGER_MATRIX &);
  friend VOID Initialize (INTEGER_MATRIX &, INT);
  friend INTEGER_MATRIX operator + (CONST INTEGER_MATRIX &, CONST INTEGER_MATRIX &);
  friend INTEGER_MATRIX operator - (CONST INTEGER_MATRIX &, CONST INTEGER_MATRIX &);
  friend INTEGER_MATRIX operator * (INT, CONST INTEGER_MATRIX &);
  friend INTEGER_MATRIX operator / (CONST INTEGER_MATRIX &, INT);
  friend INTEGER_VECTOR operator * (CONST INTEGER_MATRIX &, CONST INTEGER_VECTOR &);
  friend ostream & operator << (ostream &, CONST INTEGER_MATRIX &);
  friend istream & operator >> (istream &, INTEGER_MATRIX &);
};

#endif /* __INTEGER_MATRIX__ */
