/************************************************************************
 *
 * Definition of type INTEGER_VECTOR
 * ---------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: IntegerVector.h,v 1.2 1999/07/28 08:56:35 knueppel Exp $
 *
 ************************************************************************/

#ifndef __INTEGER_VECTOR__
#define __INTEGER_VECTOR__

#include <Configuration.h>
#include <Error.h>
#include <iostream.h>

class INTEGER_VECTOR {
  INT nElements;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  INT *theElements;
  INTEGER_VECTOR ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nElements = 0; theElements = NULL; }
  INTEGER_VECTOR (INT n)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckVectorRange (n);
      nElements = n;
      theElements = new INT[n];
      CheckMemory (theElements);
    }
  ~INTEGER_VECTOR ()
    { if (theElements != NULL) delete theElements; }
  INTEGER_VECTOR (CONST INTEGER_VECTOR &);
  INT & operator () (INT i) CONST
    { CheckVectorIndex (i, *this); return theElements[i-1]; }
  INTEGER_VECTOR & operator =  (CONST INTEGER_VECTOR &);
  INTEGER_VECTOR & operator += (CONST INTEGER_VECTOR &);
  INTEGER_VECTOR & operator -= (CONST INTEGER_VECTOR &);
  INTEGER_VECTOR & operator *= (INT);
  INTEGER_VECTOR & operator /= (INT);
  friend INT Dimension (CONST INTEGER_VECTOR & x) { return x.nElements; }
  friend VOID Resize (INTEGER_VECTOR &, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (INTEGER_VECTOR & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (INTEGER_VECTOR & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (INTEGER_VECTOR & x) { }
  friend VOID MakePermanent (INTEGER_VECTOR & x) { }
#endif
  friend VOID Clear (INTEGER_VECTOR &);
  friend VOID Initialize (INTEGER_VECTOR &, INT);
  friend INTEGER_VECTOR operator + (CONST INTEGER_VECTOR &, CONST INTEGER_VECTOR &);
  friend INTEGER_VECTOR operator - (CONST INTEGER_VECTOR &, CONST INTEGER_VECTOR &);
  friend INTEGER_VECTOR operator * (INT                   , CONST INTEGER_VECTOR &);
  friend INTEGER_VECTOR operator / (CONST INTEGER_VECTOR &, INT                   );
  friend INT            operator * (CONST INTEGER_VECTOR &, CONST INTEGER_VECTOR &);
  friend ostream & operator << (ostream &, CONST INTEGER_VECTOR &);
  friend istream & operator >> (istream &, INTEGER_VECTOR &);
};

#endif /* __INTEGER_VECTOR__ */
