/************************************************************************
 *
 * Definition of type INTERVAL
 * ---------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: Interval.h,v 1.1 1997/09/23 11:50:53 knueppel Exp $
 *
 ************************************************************************/

#ifndef __INTERVAL__
#define __INTERVAL__

#include <BIAS/Bias0.h>
#include <Configuration.h>
#include <iostream.h>

class INTERVAL {
  BIASINTERVAL ival;
public:
  INTERVAL () { }
  INTERVAL (REAL a)
    { BiasHullR(& ival, & a); }
  INTERVAL (REAL a, REAL b)
    { BiasHullRR(& ival, & a, & b); }
  // Because the appropiate routines are missing in Bias, a temporary
  // variable is needed for +=, -=, *=, /=.
  INTERVAL & operator += (REAL x)
    { INTERVAL t;
      BiasAddIR (Bias(t), Bias(*this), & x);
      *this = t; return *this;
    }
  INTERVAL & operator -= (REAL x)
    { INTERVAL t;
      BiasSubIR (Bias(t), Bias(*this), & x);
      *this = t; return *this;
    }
  INTERVAL & operator *= (REAL x)
    { INTERVAL t;
      BiasMulIR (Bias(t), Bias(*this), & x);
      *this = t; return *this;
    }
  INTERVAL & operator /= (REAL x)
    { INTERVAL t;
      BiasDivIR (Bias(t), Bias(*this), & x);
      *this = t; return *this;
    }
  INTERVAL & operator += (CONST INTERVAL & x)
    { INTERVAL t;
      BiasAddII (Bias(t), Bias(*this), Bias(x));
      *this = t; return *this;
    }
  INTERVAL & operator -= (CONST INTERVAL & x)
    { INTERVAL t;
      BiasSubII (Bias(t), Bias(*this), Bias(x));
      *this = t; return *this;
    }
  INTERVAL & operator *= (CONST INTERVAL & x)
    { INTERVAL t;
      BiasMulII (Bias(t), Bias(*this), Bias(x));
      *this = t; return *this;
    }
  INTERVAL & operator /= (CONST INTERVAL & x)
    { INTERVAL t;
      BiasDivII (Bias(t), Bias(*this), Bias(x));
      *this = t; return *this;
    }
  friend BIASINTERVAL * Bias (CONST INTERVAL & x)
    { return (BIASINTERVAL *) &(x.ival); }
  friend INTERVAL operator + (CONST INTERVAL & x) { return x; }
  friend INTERVAL operator - (CONST INTERVAL & x)
    { INTERVAL result;
      BiasNeg (Bias(result), Bias(x));
      return result;
    }
  friend INTERVAL AddBounds (REAL x, REAL y)
    { INTERVAL result;
      BiasAddRR (Bias(result), & x, & y);
      return result;
    }
  friend INTERVAL operator + (REAL x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasAddRI (Bias(result), & x, Bias(y));
      return result;
    }
  friend INTERVAL operator + (CONST INTERVAL & x, REAL y)
    { INTERVAL result;
      BiasAddIR (Bias(result), Bias(x), & y);
      return result;
    }
  friend INTERVAL operator + (CONST INTERVAL & x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasAddII (Bias(result), Bias(x), Bias(y));
      return result;
    }
  friend INTERVAL SubBounds (REAL x, REAL y)
    { INTERVAL result;
      BiasSubRR (Bias(result), & x, & y);
      return result;
    }
  friend INTERVAL operator - (REAL x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasSubRI (Bias(result), & x, Bias(y));
      return result;
    }
  friend INTERVAL operator - (CONST INTERVAL & x, REAL y)
    { INTERVAL result;
      BiasSubIR (Bias(result), Bias(x), & y);
      return result;
    }
  friend INTERVAL operator - (CONST INTERVAL & x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasSubII (Bias(result), Bias(x), Bias(y));
      return result;
    }
  friend INTERVAL MulBounds (REAL x, REAL y)
    { INTERVAL result;
      BiasMulRR (Bias(result), & x, & y);
      return result;
    }
  friend INTERVAL operator * (REAL x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasMulRI (Bias(result), & x, Bias(y));
      return result;
    }
  friend INTERVAL operator * (CONST INTERVAL & x, REAL y)
    { INTERVAL result;
      BiasMulIR (Bias(result), Bias(x), & y);
      return result;
    }
  friend INTERVAL operator * (CONST INTERVAL & x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasMulII (Bias(result), Bias(x), Bias(y));
      return result;
    }
  friend INTERVAL DivBounds (REAL x, REAL y)
    { INTERVAL result;
      BiasDivRR (Bias(result), & x, & y);
      return result;
    }
  friend INTERVAL operator / (REAL x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasDivRI (Bias(result), & x, Bias(y));
      return result;
    }
  friend INTERVAL operator / (CONST INTERVAL & x, REAL y)
    { INTERVAL result;
      BiasDivIR (Bias(result), Bias(x), & y);
      return result;
    }
  friend INTERVAL operator / (CONST INTERVAL & x, CONST INTERVAL & y)
    { INTERVAL result;
      BiasDivII (Bias(result), Bias(x), Bias(y));
      return result;
    }
  friend REAL Inf (CONST INTERVAL & x)
    { return BiasInf (Bias(x)); }
  friend REAL Sup (CONST INTERVAL & x)
    { return BiasSup (Bias(x)); }
  friend REAL Pred (REAL a) { return BiasPredR (& a); }
  friend REAL Succ (REAL a) { return BiasSuccR (& a); }
  friend INTERVAL Pred (CONST INTERVAL & x)
    { INTERVAL result;
      BiasPredI (Bias(result), Bias(x));
      return result;
    }
  friend INTERVAL Succ (CONST INTERVAL & x)
    { INTERVAL result;
      BiasSuccI (Bias(result), Bias(x));
      return result;
    }
  friend INTERVAL Hull (REAL a)
    { INTERVAL result;
      BiasHullR (Bias(result), & a);
      return result;
    }
  friend INTERVAL Hull (REAL a, REAL b)
    { INTERVAL result;
      BiasHullRR (Bias(result), & a, & b);
      return result;
    }
  friend INTERVAL Hull (REAL a, CONST INTERVAL & b)
    { INTERVAL result;
      BiasHullRI (Bias(result), & a, Bias(b));
      return result;
    }
  friend INTERVAL Hull (CONST INTERVAL & a, REAL b)
    { INTERVAL result;
      BiasHullIR (Bias(result), Bias(a), & b);
      return result;
    }
  friend INTERVAL Hull (CONST INTERVAL & a, CONST INTERVAL & b)
    { INTERVAL result;
      BiasHullII (Bias(result), Bias(a), Bias(b));
      return result;
    }
  friend INTERVAL SymHull (REAL a)
    { INTERVAL result;
      REAL minus_a = -a;
      BiasHullRR (Bias(result), & a, & minus_a);
      return result;
    }
  friend REAL Mid (CONST INTERVAL & a) { return BiasMid (Bias(a)); }
  friend REAL Diam (CONST INTERVAL & a) { return BiasDiam (Bias(a)); }
  friend REAL Abs (CONST INTERVAL & a) { return BiasAbs (Bias(a)); }
  friend REAL Mig (CONST INTERVAL & a) { return Distance (0.0, a); }
  friend REAL Distance (REAL a, CONST INTERVAL & b)
    { return BiasDistRI (& a, Bias(b)); }
  friend REAL Distance (CONST INTERVAL & a, REAL b)
    { return BiasDistIR (Bias(a), & b); }
  friend REAL Distance (CONST INTERVAL & a, CONST INTERVAL & b)
    { return BiasDistII (Bias(a), Bias(b)); }
  friend INT Intersection (INTERVAL & r, CONST INTERVAL & a, CONST INTERVAL & b)
    { return BiasIntersection (Bias(r), Bias(a), Bias(b)); }
  friend INT operator <= (REAL a, CONST INTERVAL & b)
    { return BiasInR (& a, Bias(b)); }
  friend INT operator <= (CONST INTERVAL & a, CONST INTERVAL & b)
    { return BiasInI (Bias(a), Bias(b)); }
  friend INT operator < (REAL a, CONST INTERVAL & b)
    { return BiasInInteriorR (& a, Bias(b)); }
  friend INT operator < (CONST INTERVAL & a, CONST INTERVAL & b)
    { return BiasInInteriorI (Bias(a), Bias(b)); }
  friend INT operator == (CONST INTERVAL & a, CONST INTERVAL & b)
    { return BiasIsEqual (Bias(a), Bias(b)); }
  friend INT operator != (CONST INTERVAL & a, CONST INTERVAL & b)
    { return !BiasIsEqual (Bias(a), Bias(b)); }
  friend ostream & operator << (ostream &, CONST INTERVAL &);
  friend istream & operator >> (istream &, INTERVAL &);
};

inline INTERVAL operator ++ (INTERVAL & x) { return Succ (x); }
inline INTERVAL operator -- (INTERVAL & x) { return Pred (x); }

#endif /* __INTERVAL__ */
