/************************************************************************
 *
 * Definition of type INTERVAL_MATRIX
 * ----------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: IntervalMatrix.h,v 1.2 1999/07/28 08:55:44 knueppel Exp $
 *
 ************************************************************************/

#ifndef __INTERVAL_MATRIX__
#define __INTERVAL_MATRIX__

#include <Matrix.h>
#include <Error.h>
#include <IntervalVector.h>

class INTERVAL_MATRIX {
  INT nRows;
  INT nCols;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  BIASINTERVAL *theElements;
  INTERVAL_MATRIX ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nRows = nCols = 0; theElements = NULL; }
  INTERVAL_MATRIX (INT r, INT c)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckMatrixRange (r, c);
      nRows = r; nCols = c;
      theElements = new BIASINTERVAL[r * c];
      CheckMemory (theElements);
    }
  ~INTERVAL_MATRIX ()
    { if (theElements != NULL) delete theElements; }
  INTERVAL_MATRIX (CONST INTERVAL_MATRIX &);
  INTERVAL & operator () (INT i, INT j) CONST
    { CheckMatrixIndex (i, j, *this); return * ((INTERVAL *)(theElements + (i-1) * nCols + (j-1))); }
  INTERVAL_MATRIX & operator =  (CONST INTERVAL_MATRIX &);
  INTERVAL_MATRIX & operator =  (CONST MATRIX &);
  INTERVAL_MATRIX & operator += (CONST INTERVAL_MATRIX &);
  INTERVAL_MATRIX & operator -= (CONST INTERVAL_MATRIX &);
  INTERVAL_MATRIX & operator *= (CONST INTERVAL &);
  INTERVAL_MATRIX & operator /= (CONST INTERVAL &);
  INTERVAL_MATRIX & operator += (CONST MATRIX &);
  INTERVAL_MATRIX & operator -= (CONST MATRIX &);
  INTERVAL_MATRIX & operator *= (REAL);
  INTERVAL_MATRIX & operator /= (REAL);
  friend INT RowDimension (CONST INTERVAL_MATRIX & x) { return x.nRows; }
  friend INT ColDimension (CONST INTERVAL_MATRIX & x) { return x.nCols; }
  friend VOID Resize (INTERVAL_MATRIX &, INT, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (INTERVAL_MATRIX & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (INTERVAL_MATRIX & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (INTERVAL_MATRIX & x) { }
  friend VOID MakePermanent (INTERVAL_MATRIX & x) { }
#endif
  friend VOID Clear (INTERVAL_MATRIX &);
  friend VOID Initialize (INTERVAL_MATRIX &, CONST INTERVAL &);
  friend INTERVAL_VECTOR Row (CONST INTERVAL_MATRIX &, INT);
  friend INTERVAL_VECTOR Col (CONST INTERVAL_MATRIX &, INT);
  friend VOID SetRow (INTERVAL_MATRIX &, INT, CONST INTERVAL_VECTOR &);
  friend VOID SetCol (INTERVAL_MATRIX &, INT, CONST INTERVAL_VECTOR &);
  friend MATRIX Inf (CONST INTERVAL_MATRIX &);
  friend MATRIX Sup (CONST INTERVAL_MATRIX &);
  friend MATRIX Mid (CONST INTERVAL_MATRIX &);
  friend MATRIX Diam (CONST INTERVAL_MATRIX &);
  friend MATRIX Abs (CONST INTERVAL_MATRIX &);
  friend INT Intersection (INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator <= (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator <= (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator <  (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator <  (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator == (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INT operator != (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX Hull (CONST MATRIX &);
  friend INTERVAL_MATRIX Hull (CONST MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX Hull (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX Hull (CONST INTERVAL_MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX Hull (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX SymHull (CONST MATRIX & x) { return Hull (-x, x); }
  friend INTERVAL_MATRIX operator + (CONST INTERVAL_MATRIX & a) { return (INTERVAL_MATRIX &) a; }
  friend INTERVAL_MATRIX operator - (CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX operator + (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX operator + (CONST INTERVAL_MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator + (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX AddBounds  (CONST MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator - (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX operator - (CONST INTERVAL_MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator - (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX SubBounds  (CONST MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator * (REAL, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX operator * (CONST INTERVAL &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator * (CONST INTERVAL &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX MulBounds  (REAL, CONST MATRIX &);
  friend INTERVAL_MATRIX operator / (CONST INTERVAL_MATRIX &, REAL);
  friend INTERVAL_MATRIX operator / (CONST MATRIX &, CONST INTERVAL &);
  friend INTERVAL_MATRIX operator / (CONST INTERVAL_MATRIX &, CONST INTERVAL &);
  friend INTERVAL_MATRIX DivBounds  (CONST MATRIX &, REAL);
  friend INTERVAL_VECTOR operator * (CONST MATRIX &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR operator * (CONST INTERVAL_MATRIX &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator * (CONST INTERVAL_MATRIX &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR MulBounds  (CONST MATRIX &, CONST VECTOR &);
  friend INTERVAL_MATRIX operator * (CONST MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX operator * (CONST INTERVAL_MATRIX &, CONST MATRIX &);
  friend INTERVAL_MATRIX operator * (CONST INTERVAL_MATRIX &, CONST INTERVAL_MATRIX &);
  friend INTERVAL_MATRIX MulBounds  (CONST MATRIX &, CONST MATRIX &);
  friend ostream & operator << (ostream &, CONST INTERVAL_MATRIX &);
  friend istream & operator >> (istream &, INTERVAL_MATRIX &);
};

#endif /* __INTERVAL_MATRIX__ */
