/************************************************************************
 *
 * Definition of type INTERVAL_VECTOR
 * ----------------------------------
 *
 * Copyright (c) 1993, 1995, 1997 Olaf Knueppel
 *
 * $Id: IntervalVector.h,v 1.2 1999/07/28 08:56:37 knueppel Exp $
 *
 ************************************************************************/

#ifndef __INTERVAL_VECTOR__
#define __INTERVAL_VECTOR__

#include <BIAS/Bias1.h>
#include <Configuration.h>
#include <Error.h>
#include <Vector.h>
#include <Interval.h>

class INTERVAL_VECTOR {
  INT nElements;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  BIASINTERVAL *theElements;
  INTERVAL_VECTOR ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nElements = 0; theElements = NULL; }
  INTERVAL_VECTOR (INT n)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckVectorRange (n);
      nElements = n;
      theElements = new BIASINTERVAL[n];
      CheckMemory (theElements);
    }
  ~INTERVAL_VECTOR ()
    { if (theElements != NULL) delete theElements; }
  INTERVAL_VECTOR (CONST INTERVAL_VECTOR &);
  INTERVAL & operator () (INT i) CONST
    { CheckVectorIndex (i, *this); return * ((INTERVAL *)(theElements + i-1)); }
  INTERVAL_VECTOR & operator =  (CONST INTERVAL_VECTOR &);
  INTERVAL_VECTOR & operator =  (CONST VECTOR &);
  INTERVAL_VECTOR & operator += (CONST INTERVAL_VECTOR &);
  INTERVAL_VECTOR & operator -= (CONST INTERVAL_VECTOR &);
  INTERVAL_VECTOR & operator *= (CONST INTERVAL &);
  INTERVAL_VECTOR & operator /= (CONST INTERVAL &);
  INTERVAL_VECTOR & operator += (CONST VECTOR &);
  INTERVAL_VECTOR & operator -= (CONST VECTOR &);
  INTERVAL_VECTOR & operator *= (REAL);
  INTERVAL_VECTOR & operator /= (REAL);
  friend INT Dimension (CONST INTERVAL_VECTOR & x) { return x.nElements; }
  friend VOID Resize (INTERVAL_VECTOR &, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (INTERVAL_VECTOR & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (INTERVAL_VECTOR & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (INTERVAL_VECTOR & x) { }
  friend VOID MakePermanent (INTERVAL_VECTOR & x) { }
#endif
  friend VOID Clear (INTERVAL_VECTOR &);
  friend VOID Initialize (INTERVAL_VECTOR &, CONST INTERVAL &);
  friend VECTOR Inf (CONST INTERVAL_VECTOR &);
  friend VECTOR Sup (CONST INTERVAL_VECTOR &);
  friend VECTOR Mid (CONST INTERVAL_VECTOR &);
  friend VECTOR Diam (CONST INTERVAL_VECTOR &);
  friend VECTOR Abs (CONST INTERVAL_VECTOR &);
  friend INT Intersection (INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator <= (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator <= (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator <  (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator <  (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator == (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INT operator != (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR Hull (CONST VECTOR &);
  friend INTERVAL_VECTOR Hull (CONST VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR Hull (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR Hull (CONST INTERVAL_VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR Hull (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR SymHull (CONST VECTOR & x) { return Hull (-x, x); }
  friend INTERVAL_VECTOR operator + (CONST INTERVAL_VECTOR & a) { return (INTERVAL_VECTOR &) a; }
  friend INTERVAL_VECTOR operator - (CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR operator + (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR operator + (CONST INTERVAL_VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator + (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR AddBounds  (CONST VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator - (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR operator - (CONST INTERVAL_VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator - (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR SubBounds  (CONST VECTOR &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator * (REAL, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR operator * (CONST INTERVAL &, CONST VECTOR &);
  friend INTERVAL_VECTOR operator * (CONST INTERVAL &, CONST INTERVAL_VECTOR &);
  friend INTERVAL_VECTOR MulBounds  (REAL, CONST VECTOR &);
  friend INTERVAL_VECTOR operator / (CONST INTERVAL_VECTOR &, REAL);
  friend INTERVAL_VECTOR operator / (CONST VECTOR &, CONST INTERVAL &);
  friend INTERVAL_VECTOR operator / (CONST INTERVAL_VECTOR &, CONST INTERVAL &);
  friend INTERVAL_VECTOR DivBounds  (CONST VECTOR &, REAL);
  friend INTERVAL operator * (CONST VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL operator * (CONST INTERVAL_VECTOR &, CONST VECTOR &);
  friend INTERVAL operator * (CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend INTERVAL MulBounds  (CONST VECTOR &, CONST VECTOR &);
  friend INTERVAL Sqr  (CONST INTERVAL_VECTOR &);
  friend INTERVAL Norm (CONST INTERVAL_VECTOR &);
  friend ostream & operator << (ostream &, CONST INTERVAL_VECTOR &);
  friend istream & operator >> (istream &, INTERVAL_VECTOR &);
};

#endif /* __INTERVAL_VECTOR__ */
