/************************************************************************
 *
 * Implementation of Solvers for Linear Systems
 * --------------------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: LSS.C,v 1.1 1997/09/23 11:50:54 knueppel Exp $
 *
 ************************************************************************/

#include <LSS.h>
#include <Utilities.h>
#include <Constants.h>

INTERVAL_VECTOR LSS (CONST MATRIX & A, CONST VECTOR & b, INT & info)
{
  INT dim = Dimension (b);
  MATRIX R (dim, dim);
  VECTOR xs (dim);
  INTERVAL_VECTOR x (dim), y (dim), z (dim);
  INTERVAL_VECTOR Inflat (dim);
  INTERVAL_MATRIX C(dim,dim);
  INT k, done;
  INTERVAL eps (0.9,1.1);

  Initialize(Inflat, SymHull (Machine::MinPositive));
  
  R = Inverse (A);
  xs = R * b;

  z = R * (b - MulBounds (A, xs));
  
  x = z;
  
  C = Id(dim) - MulBounds (R, A);

  k = 0;
  do {
    y = eps * x + Inflat; // expansion
    x = z + C * y;
    done = (x < y);
    k++;
  } while (!done && k < 10); 

  info = done;              // 1, if inclusion was possible
  return (xs + x);
}

INTERVAL_VECTOR ILSS (CONST INTERVAL_MATRIX & A, CONST INTERVAL_VECTOR & b,
                      INT & info)
{
  INT dim = Dimension (b);
  MATRIX R (dim, dim);
  VECTOR xs (dim);
  INTERVAL_VECTOR x (dim), y (dim), z (dim);
  INTERVAL_VECTOR Inflat (dim);
  INTERVAL_MATRIX C(dim,dim);
  INT k, done;
  INTERVAL eps (0.9,1.1);

  Initialize(Inflat, SymHull (Machine::MinPositive));
  
  R = Inverse (Mid (A));
  xs = R * Mid (b);

  z = R * (b - A * xs);

  x = z;
  
  C = Id(dim) - R * A;

  k = 0;
  do {
    y = eps * x + Inflat; // expansion
    x = z + C * y;
    done = (x < y);
    k++;
  } while (!done && k < 10); 

  info = done;              // 1, if inclusion was possible
  return (xs + x);
}
