/************************************************************************
 *
 * Definition of type LONGINTERVAL
 * -------------------------------
 *
 * Copyright (c) 1993, 1997 Olaf Knueppel
 *
 * $Id: LongInterval.h,v 1.2 1999/07/28 08:55:27 knueppel Exp $
 *
 ************************************************************************/

#ifndef __LONGINTERVAL__
#define __LONGINTERVAL__

#include <Configuration.h>
#include <LongReal.h>
#include <IntervalVector.h>
#include <lr/Li.h>
#include <lr/DynMem.h>

class LONGINTERVAL {
  static USHORT OutOptions;           // output options (long intervals only)
  USHORT nDigits;                     // total number of digits (long real base)
  USHORT nGuards;                     // number of guard digits (long real base)
#ifdef __DONTCOPY__
  INT    isTemporary;                 // set, if variable is only temporary
#endif
  LIREAL  value;                      // pointer to long real number
  LONGINTERVAL (INT);                 // private to avoid typecast INT -> LONGINTERVAL
public:
  LONGINTERVAL ();
  LONGINTERVAL (REAL);
  LONGINTERVAL (CONST INTERVAL &);
  LONGINTERVAL (CONST LONGREAL &);
  LONGINTERVAL (CONST LONGINTERVAL &);
  LONGINTERVAL (CONST CHAR *, INT n = 0);
  ~LONGINTERVAL ();
  LONGINTERVAL & operator = (REAL);
  LONGINTERVAL & operator = (CONST INTERVAL &);
  LONGINTERVAL & operator = (CONST LONGINTERVAL &);
  LONGINTERVAL & operator += (REAL);
  LONGINTERVAL & operator += (CONST INTERVAL &);
  LONGINTERVAL & operator += (CONST LONGINTERVAL &);
  LONGINTERVAL & operator -= (REAL);
  LONGINTERVAL & operator -= (CONST INTERVAL &);
  LONGINTERVAL & operator -= (CONST LONGINTERVAL &);
  LONGINTERVAL & operator *= (REAL);
  LONGINTERVAL & operator *= (CONST INTERVAL &);
  LONGINTERVAL & operator *= (CONST LONGINTERVAL &);
  LONGINTERVAL & operator /= (REAL);
  LONGINTERVAL & operator /= (CONST INTERVAL &);
  LONGINTERVAL & operator /= (CONST LONGINTERVAL &);
  friend LONGINTERVAL Hull (CONST LONGREAL &);
  friend LONGINTERVAL Hull (CONST LONGREAL &, CONST LONGREAL &);
  friend LONGREAL Inf (CONST LONGINTERVAL &);
  friend LONGREAL Sup (CONST LONGINTERVAL &);
  friend LONGREAL Diam (CONST LONGINTERVAL &);
  friend LONGREAL Mid (CONST LONGINTERVAL &);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (LONGINTERVAL & x) { x.isTemporary = 1; }
  friend VOID MakePermanent (LONGINTERVAL & x) { x.isTemporary = 0; }
#else
  friend VOID MakeTemporary (LONGINTERVAL & x) { }
  friend VOID MakePermanent (LONGINTERVAL & x) { }
#endif
  friend LIREAL Value (CONST LONGINTERVAL & x) { return x.value; }
  friend INTERVAL RoundToInterval (CONST LONGINTERVAL &);
  friend LONGINTERVAL LongIntervalEnclosure (CONST CHAR *);
  friend VOID Accumulate (LONGINTERVAL &, CONST VECTOR &         , CONST VECTOR &);
  friend VOID Accumulate (LONGINTERVAL &, CONST VECTOR &         , CONST INTERVAL_VECTOR &);
  friend VOID Accumulate (LONGINTERVAL &, CONST INTERVAL_VECTOR &, CONST VECTOR &);
  friend VOID Accumulate (LONGINTERVAL &, CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &);
  friend LONGINTERVAL operator - (CONST LONGINTERVAL &);
  friend LONGINTERVAL operator + (CONST LONGINTERVAL &, REAL);
  friend LONGINTERVAL operator + (CONST LONGINTERVAL &, CONST INTERVAL &);
  friend LONGINTERVAL operator + (REAL,                 CONST LONGINTERVAL &);
  friend LONGINTERVAL operator + (CONST INTERVAL &,     CONST LONGINTERVAL &);
  friend LONGINTERVAL operator + (CONST LONGINTERVAL &, CONST LONGINTERVAL &);
  friend LONGINTERVAL operator - (CONST LONGINTERVAL &, REAL);
  friend LONGINTERVAL operator - (CONST LONGINTERVAL &, CONST INTERVAL &);
  friend LONGINTERVAL operator - (REAL,                 CONST LONGINTERVAL &);
  friend LONGINTERVAL operator - (CONST INTERVAL &,     CONST LONGINTERVAL &);
  friend LONGINTERVAL operator - (CONST LONGINTERVAL &, CONST LONGINTERVAL &);
  friend LONGINTERVAL operator * (CONST LONGINTERVAL &, REAL);
  friend LONGINTERVAL operator * (CONST LONGINTERVAL &, CONST INTERVAL &);
  friend LONGINTERVAL operator * (REAL,                 CONST LONGINTERVAL &);
  friend LONGINTERVAL operator * (CONST INTERVAL &,     CONST LONGINTERVAL &);
  friend LONGINTERVAL operator * (CONST LONGINTERVAL &, CONST LONGINTERVAL &);
  friend LONGINTERVAL operator / (CONST LONGINTERVAL &, REAL);
  friend LONGINTERVAL operator / (CONST LONGINTERVAL &, CONST INTERVAL &);
  friend LONGINTERVAL operator / (REAL,                 CONST LONGINTERVAL &);
  friend LONGINTERVAL operator / (CONST INTERVAL &,     CONST LONGINTERVAL &);
  friend LONGINTERVAL operator / (CONST LONGINTERVAL &, CONST LONGINTERVAL &);
  friend ostream & operator << (ostream &, CONST LONGINTERVAL &);
  friend istream & operator >> (istream &, LONGINTERVAL &);
  friend INT GetOutOptions ();
  friend VOID SetOutOptions (INT);
};

VOID LiToOutput (ostream &, LIREAL);
VOID StringToLi (LIREAL, CONST CHAR *);
INTERVAL Enclosure (CONST CHAR *);

#endif /* __LONGINTERVAL__ */
