/************************************************************************
 *
 * Definition of type LONGREAL
 * ---------------------------
 *
 * Copyright (c) 1993, 1994, 1997 Olaf Knueppel
 *
 * $Id: LongReal.h,v 1.2 1999/07/28 08:55:29 knueppel Exp $
 *
 ************************************************************************/

#ifndef __LONGREAL__
#define __LONGREAL__

#include <Configuration.h>
#include <Vector.h>
#include <lr/Lr.h>
#include <lr/DynMem.h>

class LONGREAL {
  static USHORT nCurrentDigits;   // current total number of digits
  static USHORT nOutFracDigits;   // total number of fraction digits for output
  static USHORT nOutIntDigits;    // total number of integer digits for output
                                  // (long intervals only)
  static INT OutRndMode;          // rounding mode for output

  USHORT nDigits;                 // total number of digits (long real base)
  USHORT nGuards;                 // number of guard digits (long real base)
#ifdef __DONTCOPY__
  INT    isTemporary;             // set, if variable is only temporary
#endif
  LREAL  value;                   // pointer to long real number
  LONGREAL (INT);                 // private to avoid typecast INT -> LONGREAL
public:
  LONGREAL ();
  LONGREAL (REAL);
  LONGREAL (CONST LONGREAL &);
  LONGREAL (CONST char *, INT n = 0);
  ~LONGREAL ();
  LONGREAL & operator = (CONST LONGREAL &);
  LONGREAL & operator = (REAL);
  LONGREAL & operator += (CONST LONGREAL &);
  LONGREAL & operator += (REAL);
  LONGREAL & operator -= (CONST LONGREAL &);
  LONGREAL & operator -= (REAL);
  LONGREAL & operator *= (CONST LONGREAL &);
  LONGREAL & operator *= (REAL);
  LONGREAL & operator /= (CONST LONGREAL &);
  LONGREAL & operator /= (REAL);
  LONGREAL & operator ++ ();
  LONGREAL & operator -- ();
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (LONGREAL & x) { x.isTemporary = 1; }
  friend VOID MakePermanent (LONGREAL & x) { x.isTemporary = 0; }
#else
  friend VOID MakeTemporary (LONGREAL & x) { }
  friend VOID MakePermanent (LONGREAL & x) { }
#endif
  friend LREAL Value (CONST LONGREAL & x) { return x.value; }
  friend REAL RoundToReal (CONST LONGREAL & x, INT RndMode = LR_RND_NEAR);
  friend VOID Accumulate (LONGREAL &, CONST VECTOR &, CONST VECTOR &);
  friend LONGREAL StringToLongReal (CONST CHAR *);
  friend LONGREAL operator - (CONST LONGREAL &);
  friend LONGREAL operator + (CONST LONGREAL &, REAL);
  friend LONGREAL operator + (REAL,             CONST LONGREAL &);
  friend LONGREAL operator + (CONST LONGREAL &, CONST LONGREAL &);
  friend LONGREAL operator - (CONST LONGREAL &, REAL);
  friend LONGREAL operator - (REAL,             CONST LONGREAL &);
  friend LONGREAL operator - (CONST LONGREAL &, CONST LONGREAL &);
  friend LONGREAL operator * (CONST LONGREAL &, REAL);
  friend LONGREAL operator * (REAL,             CONST LONGREAL &);
  friend LONGREAL operator * (CONST LONGREAL &, CONST LONGREAL &);
  friend LONGREAL operator / (CONST LONGREAL &, REAL);
  friend LONGREAL operator / (REAL,             CONST LONGREAL &);
  friend LONGREAL operator / (CONST LONGREAL &, CONST LONGREAL &);
  friend INT operator == (CONST LONGREAL & a, CONST LONGREAL & b)
    { return (LrCmp (Value(a), Value(b)) == LR_EQ); }
  friend INT operator != (CONST LONGREAL & a, CONST LONGREAL & b)
    { return (LrCmp (Value(a), Value(b)) != LR_EQ); }
  friend INT operator <  (CONST LONGREAL & a, CONST LONGREAL & b)
    { return (LrCmp (Value(a), Value(b)) == LR_LT); }
  friend INT operator <= (CONST LONGREAL & a, CONST LONGREAL & b)
    {
      INT c;
      c =  LrCmp (Value(a), Value(b));
      return (c == LR_LT) || (c == LR_EQ);
    }
  friend INT operator >  (CONST LONGREAL & a, CONST LONGREAL & b)
    { return (LrCmp (Value(a), Value(b)) == LR_GT); }
  friend INT operator >= (CONST LONGREAL & a, CONST LONGREAL & b)
    {
      INT c;
      c =  LrCmp (Value(a), Value(b));
      return (c == LR_GT) || (c == LR_EQ);
    }
  friend INT GetDigits ();
  friend INT GetBaseDigits ();
  friend INT GetOutFracDigits ();
  friend INT GetOutIntDigits ();
  friend INT GetOutRndMode ();
  friend VOID SetDigits (INT);
  friend VOID SetBaseDigits (INT);
  friend VOID SetOutFracDigits (INT);
  friend VOID SetOutIntDigits (INT);
  friend VOID SetOutRndMode (INT);
  friend istream & operator >> (istream &, LONGREAL &);
  friend ostream & operator << (ostream &, CONST LONGREAL &);
};

VOID StringToLr (LREAL, CONST CHAR *);

#endif /* __LONGREAL__ */
