/************************************************************************
 *
 * Definition of type MATRIX
 * -------------------------
 *
 * Copyright (c) 1993, 1995, 1997 Olaf Knueppel
 *
 * $Id: Matrix.h,v 1.2 1999/07/28 08:56:39 knueppel Exp $
 *
 ************************************************************************/

#ifndef __MATRIX__
#define __MATRIX__

#include <Configuration.h>
#include <Error.h>
#include <Vector.h>

class MATRIX {
  INT nRows;
  INT nCols;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  REAL *theElements;
  MATRIX ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nRows = nCols = 0; theElements = NULL; }
  MATRIX (INT r, INT c)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckMatrixRange (r, c);
      nRows = r; nCols = c;
      theElements = new REAL[r * c];
      CheckMemory (theElements);
    }
  ~MATRIX ()
    { if (theElements != NULL) delete theElements; }
  MATRIX (CONST MATRIX &);
  REAL & operator () (INT i, INT j) CONST
    { CheckMatrixIndex (i, j, *this); return theElements[(i-1) * nCols + (j-1)]; }
  MATRIX & operator =  (CONST MATRIX &);
  MATRIX & operator += (CONST MATRIX &);
  MATRIX & operator -= (CONST MATRIX &);
  MATRIX & operator *= (REAL);
  MATRIX & operator /= (REAL);
  friend INT RowDimension (CONST MATRIX & x) { return x.nRows; }
  friend INT ColDimension (CONST MATRIX & x) { return x.nCols; }
  friend VOID Resize (MATRIX &, INT, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (MATRIX & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (MATRIX & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (MATRIX & x) { }
  friend VOID MakePermanent (MATRIX & x) { }
#endif
  friend VOID Clear (MATRIX &);
  friend VOID Initialize (MATRIX &, REAL);
  friend VECTOR Row (CONST MATRIX &, INT);
  friend VECTOR Col (CONST MATRIX &, INT);
  friend VOID SetRow (MATRIX &, INT, CONST VECTOR &);
  friend VOID SetCol (MATRIX &, INT, CONST VECTOR &);
  friend MATRIX operator + (CONST MATRIX & a) { return (MATRIX &) a; }
  friend MATRIX operator - (CONST MATRIX &);
  friend MATRIX operator + (CONST MATRIX &, CONST MATRIX &);
  friend MATRIX operator - (CONST MATRIX &, CONST MATRIX &);
  friend MATRIX operator * (REAL          , CONST MATRIX &);
  friend MATRIX operator / (CONST MATRIX &, REAL          );
  friend VECTOR operator * (CONST MATRIX &, CONST VECTOR &);
  friend MATRIX operator * (CONST MATRIX &, CONST MATRIX &);
  friend INT    operator <  (CONST MATRIX &, CONST MATRIX &);
  friend INT    operator <= (CONST MATRIX &, CONST MATRIX &);
  friend INT    operator >  (CONST MATRIX &, CONST MATRIX &);
  friend INT    operator >= (CONST MATRIX &, CONST MATRIX &);
  friend MATRIX Abs (CONST MATRIX &);
  friend ostream & operator << (ostream &, CONST MATRIX &);
  friend istream & operator >> (istream &, MATRIX &);
};

#endif /* __MATRIX__ */
