/************************************************************************
 *
 * Definition of type VECTOR
 * -------------------------
 *
 * Copyright (c) 1993, 1994, 1997 Olaf Knueppel
 *
 * $Id: Vector.h,v 1.2 1999/07/28 08:56:41 knueppel Exp $
 *
 ************************************************************************/

#ifndef __VECTOR__
#define __VECTOR__

#include <Configuration.h>
#include <Error.h>
#include <iostream.h>

class VECTOR {
  INT nElements;
#ifdef __DONTCOPY__
  INT IsTemporary;
#endif
public:
  REAL *theElements;
  VECTOR ()
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      nElements = 0; theElements = NULL; }
  VECTOR (INT n)
    {
#ifdef __DONTCOPY__
      IsTemporary = 0;
#endif
      CheckVectorRange (n);
      nElements = n;
      theElements = new REAL[n];
      CheckMemory (theElements);
    }
  ~VECTOR ()
    { if (theElements != NULL) delete theElements; }
  VECTOR (CONST VECTOR &);
  REAL & operator () (INT i) CONST
    { CheckVectorIndex (i, *this); return theElements[i-1]; }
  VECTOR & operator =  (CONST VECTOR &);
  VECTOR & operator += (CONST VECTOR &);
  VECTOR & operator -= (CONST VECTOR &);
  VECTOR & operator *= (REAL);
  VECTOR & operator /= (REAL);
  friend INT Dimension (CONST VECTOR & x) { return x.nElements; }
  friend VOID Resize (VECTOR &, INT);
#ifdef __DONTCOPY__
  friend VOID MakeTemporary (VECTOR & x) { x.IsTemporary = 1; }
  friend VOID MakePermanent (VECTOR & x) { x.IsTemporary = 0; }
#else
  friend VOID MakeTemporary (VECTOR & x) { }
  friend VOID MakePermanent (VECTOR & x) { }
#endif
  friend VOID Clear (VECTOR &);
  friend VOID Initialize (VECTOR &, REAL);
  friend VECTOR operator + (CONST VECTOR & a) { return (VECTOR &) a; }
  friend VECTOR operator - (CONST VECTOR &);
  friend VECTOR operator + (CONST VECTOR &, CONST VECTOR &);
  friend VECTOR operator - (CONST VECTOR &, CONST VECTOR &);
  friend VECTOR operator * (REAL    , CONST VECTOR &);
  friend VECTOR operator / (CONST VECTOR &, REAL    );
  friend REAL   operator * (CONST VECTOR &, CONST VECTOR &);
  friend INT    operator <  (CONST VECTOR &, CONST VECTOR &);
  friend INT    operator <= (CONST VECTOR &, CONST VECTOR &);
  friend INT    operator >  (CONST VECTOR &, CONST VECTOR &);
  friend INT    operator >= (CONST VECTOR &, CONST VECTOR &);
  friend VECTOR Abs  (CONST VECTOR &);
  friend REAL   Sqr  (CONST VECTOR &);
  friend REAL   Norm (CONST VECTOR &);
  friend REAL   Max  (CONST VECTOR &);
  friend REAL   Min  (CONST VECTOR &);
  friend ostream & operator << (ostream &, CONST VECTOR &);
  friend istream & operator >> (istream &, VECTOR &);
};

#endif /* __VECTOR__ */
