/************************************************************************
 *
 * Definition of Approximation List
 * --------------------------------
 *
 * Copyright (c) 1995, 1997 Olaf Knueppel
 *
 * $Id: AppList.h,v 1.2 1997/09/25 09:18:30 knueppel Exp $
 *
 ************************************************************************/

#ifndef __APPLIST__
#define __APPLIST__

#include <IntervalMatrix.h>
#include <Boolean.h>

class APPROX_ELEMENT {
 public:
  REAL            MinValue;            // approx. minimum value
  VECTOR          MinPoint;            // approx. minimum point
  INTERVAL_VECTOR HeuristicBox;        // heuristic area around minimum point
  INTERVAL_VECTOR MinEnclosure;        // enclosure of minimum point
  INTERVAL        MinValueEnclosure;   // enclosure of minimum value
  INTERVAL_VECTOR ExpansionBox;        // expansion around minimum point
  APPROX_ELEMENT() {};
  APPROX_ELEMENT(REAL &, CONST INTERVAL_VECTOR &, CONST VECTOR &);
  APPROX_ELEMENT(REAL &, CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &,
		 CONST INTERVAL_VECTOR &, CONST INTERVAL &, CONST VECTOR &);
  APPROX_ELEMENT & operator = (CONST APPROX_ELEMENT &);
  friend BOOL HasInclusion (CONST APPROX_ELEMENT & a) { return Dimension (a.MinEnclosure) != 0; }
  friend ostream & operator << (ostream &, CONST APPROX_ELEMENT &);
};

#undef LISTOBJECT
#undef LIST
#undef LIST_ELEMENT
#undef LISTCMPFUNC
#define LISTCMPFUNC  APPROXCMPFUNC
#define LIST_ELEMENT APPROX_ELEMENT
#define LIST         APPROXIMATIONLIST
#define LISTOBJECT   APPROXLISTOBJECT
#include <Lists/LinearList.hgen>

VOID Cleanup (APPROXIMATIONLIST &, REAL);
REAL Lowest (APPROXIMATIONLIST &);

VOID Append (APPROXIMATIONLIST &, CONST VECTOR &, REAL, CONST INTERVAL_VECTOR &);
VOID Append (APPROXIMATIONLIST &, CONST VECTOR &,
             REAL, CONST INTERVAL_VECTOR &,
	     CONST INTERVAL_VECTOR &, CONST INTERVAL_VECTOR &,
             CONST INTERVAL &);
VOID Insert (APPROXIMATIONLIST &, CONST VECTOR &,
             REAL, CONST INTERVAL_VECTOR &);
VOID Append_First (APPROXIMATIONLIST &, CONST VECTOR &,
                   REAL, CONST INTERVAL_VECTOR &);

VOID Append (APPROXIMATIONLIST &, APPROXIMATIONLIST &);

BOOL Approximation_Exists (APPROXIMATIONLIST &,
                           CONST INTERVAL_VECTOR &);
BOOL Approximation_Exists (APPROXIMATIONLIST &,
                           CONST INTERVAL_VECTOR &, VECTOR &);

BOOL Expansion_Exists (APPROXIMATIONLIST &,
                       CONST INTERVAL_VECTOR &);
BOOL Expansion_Exists (APPROXIMATIONLIST &,
                       CONST VECTOR &);

BOOL Approximation_In_Range (APPROXIMATIONLIST &,
                             CONST INTERVAL_VECTOR &);

#endif /* __APPLIST__ */
