/************************************************************************
 *
 * Implementation of Expansion Strategy
 * ------------------------------------
 *
 * Copyright (c) 1995, 1997 Olaf Knueppel
 *
 * $Id: Expand.C,v 1.1 1997/09/23 11:52:10 knueppel Exp $
 *
 ************************************************************************/

static const char rcs_id[] = "$Id: Expand.C,v 1.1 1997/09/23 11:52:10 knueppel Exp $";

#include <GlobalOpt/Expand.h>
#include <Utilities.h>
#include <Functions.h>
#include <LSS.h>

#define Z_SIGMA z = Sup ((Abs (-R * g) + Abs (Id (dim) - R * h) * Hull (sigma * w)));

BOOL ExpansionStep (FUNCTION & f, CONST INTERVAL_VECTOR & MaxRange,
                    CONST VECTOR & x, INTERVAL_VECTOR & Enclosure,
		    INTERVAL_VECTOR & Expansion)
{
  INT dim = Dimension (x);
  INT j;
  REAL tau = 0.001;
  REAL sigma = 0.00001;
  REAL expansion = 2;
  VECTOR w(dim), z(dim);
  INTERVAL_VECTOR g(dim);
  MATRIX R(dim, dim);
  INTERVAL_MATRIX h(dim, dim);
  INTERVAL_VECTOR x_dummy (dim), y_dummy;
  BOOL incl;
  
  for (j = 1; j <= dim; j++)
    if ((w(j) = Abs(x(j))) < tau) w(j) = 1.0;

  g = Gradient (f, Hull (x));
  h = Hessian (f, x + sigma * SymHull(w));
  ErrorHandler::SevereLevel = ErrorHandler::SevereError + 2;
  R = Inverse (Mid (h));
  ErrorHandler::SevereLevel = ErrorHandler::SevereError;
  Z_SIGMA;
  if (!(z < sigma * w)) return FALSE;
  Enclosure = x + SymHull (z);
  sigma = 0.1;
  h = Hessian (f, x + sigma * SymHull(w));
  Z_SIGMA;
  if ((z < sigma * w)) {
    do {
      Expansion = x + SymHull (sigma * w);
      sigma *= expansion;
      h = Hessian (f, x + sigma * SymHull(w));
      Z_SIGMA;
    } while ((z < sigma * w) && !(MaxRange <= Expansion));
  }
  else {
    do {
      sigma /= expansion;
      h = Hessian (f, x + sigma * SymHull(w));
      Z_SIGMA;
      Expansion = x + SymHull (sigma * w);
    } while (!(z < sigma * w) && !(Expansion <= Enclosure));
  }
  if ((Expansion <= Enclosure)) Expansion = Enclosure;

  g = x - ILSS (Hessian (f, Enclosure), g, incl);
  x_dummy = Enclosure;
  if (incl) Intersection (Enclosure, g, x_dummy);
  return TRUE;
}
