/************************************************************************
 *
 * Simple Test Program for Minimization Algorithm
 * ----------------------------------------------
 *
 * Copyright (c) 1997 Olaf Knueppel
 *
 * $Id: TestOpt.C,v 1.3 1999/07/28 08:56:28 knueppel Exp $
 *
 ************************************************************************/

#include <GlobalOpt/UnconstrainedOpt.h>
#include <Functions.h>

//
// The Branin function is used as test function.
// This function has 5 global minima.
//

INTERVAL_AUTODIFF fTestFunction (CONST INTERVAL_AUTODIFF & x)
{
  return Sqr (1.0 - 2.0 * x(2) + 0.05 * Sin (4.0 * Constant::Pi * x(2)) - x(1)) 
       + Sqr (x(2) - 0.5 * Sin (2.0 * Constant::Pi*x(1)));
}

FUNCTION TestFunction (2, fTestFunction);

VOID SetTestDomain (INTERVAL_VECTOR & x)
// The domain of the test function
{
  Resize (x, 2);
  x(1) = Hull (-10.0, 10.0);
  x(2) = Hull (-10.0, 10.0);
}


main()
{
  INT Iterations, BranchLevels;
  SOLUTIONLIST SolutionList;
  APPROXIMATIONLIST ApproximationList;
  INTERVAL_VECTOR TestDomain;
  INTERVAL ValueBounds;
  REAL LowerBound, UpperBound;

  SetTestDomain (TestDomain);

  cout << "Computing all global minima of the Branin function" << endl << endl;

  cout << "To get all minima, use e.g. nit = 6, nd = 3" << endl << endl;

  cout << "nit = "; cin >> Iterations;
  cout << "nd  = "; cin >> BranchLevels;

  StartUnconstrainedOptimization (SolutionList, ApproximationList,
				  Iterations, BranchLevels, 1e-6,
				  0.2, 0.2,
				  LowerBound, UpperBound, TestDomain,
				  TestFunction, TRUE);

  CleanUpLists (SolutionList, ApproximationList, 1e-6, LowerBound, UpperBound);
  ValueBounds = Hull (LowerBound, UpperBound);

  cout << "f min in " << ValueBounds << endl;
  cout << "Solution List:" << endl << SolutionList << endl;
  cout << "Approximation List:" << endl << ApproximationList << endl;
  cout << "Calls needed:" << endl;
  cout << "  Function calls (Real,Interval): "
       << TestFunction.RealFunctionCalls << ','
       << TestFunction.IntervalFunctionCalls << endl;
  cout << "  Gradient calls (Real,Interval): "
       << TestFunction.RealGradientCalls << ','
       << TestFunction.IntervalGradientCalls << endl;
  cout << "   Hessian calls (Real,Interval): "
       << TestFunction.RealHessianCalls << ','
       << TestFunction.IntervalHessianCalls << endl;
  return 0;
}
