/************************************************************************
 *
 * Implementation of Solution List
 * -------------------------------
 *
 * Copyright (c) 1995, 1997 Olaf Knueppel
 *
 * $Id: VecList.C,v 1.2 1997/09/25 09:18:23 knueppel Exp $
 *
 ************************************************************************/

static const char rcs_id[] = "$Id: VecList.C,v 1.2 1997/09/25 09:18:23 knueppel Exp $";

#include <GlobalOpt/VecList.h>
#include <Constants.h>
#include <Lists/LinearList.Cgen>
#undef LISTOBJECT
#undef LIST
#undef LIST_ELEMENT
#undef LISTCMPFUNC

SOLUTIONLIST_ELEMENT::SOLUTIONLIST_ELEMENT (CONST INTERVAL & va, CONST INTERVAL_VECTOR & ve)
{
  FunctionValue = va;
  Box = ve;
}

SOLUTIONLIST_ELEMENT & SOLUTIONLIST_ELEMENT::operator = (CONST SOLUTIONLIST_ELEMENT & v)
{
  FunctionValue = v.FunctionValue;
  Box = v.Box;
  return *this;
}

ostream & operator << (ostream & o, CONST SOLUTIONLIST_ELEMENT & v)
{
  return o << v.Box << "->" << v.FunctionValue;
}

BOOL SolutionCmp (CONST SOLUTIONLIST_ELEMENT & obj1,
                  CONST SOLUTIONLIST_ELEMENT & obj2)
// Compare subroutine for sorting the list by increasing lower bounds of FunctionValue
{
  return (Inf (obj1.FunctionValue) <= Inf (obj2.FunctionValue));
}

VOID Cleanup (SOLUTIONLIST & list, REAL threshold)
// Delete all list entries with Inf(F(y)) > threshold.
{
  if (IsEmpty (list)) return;
  First (list);
  while (!Finished (list))
    {
      if (Inf (Current (list).FunctionValue) > threshold) RemoveCurrent (list);
      else Next (list);
    }
}

REAL Lowest(SOLUTIONLIST& list)
// return Min{Inf(F(y))}
{
  REAL cur_min = Machine::PosInfinity;
  REAL inf;

  First (list);
  while (!Finished (list))
    {
      inf = Inf (Current (list).FunctionValue);
      if (inf < cur_min) cur_min = inf;
      Next (list);
    }
  return cur_min;
}

VOID Append (SOLUTIONLIST & list,
             CONST INTERVAL_VECTOR & vec, CONST INTERVAL & value)
{
  SOLUTIONLIST_ELEMENT iv(value, vec);
  list += iv;
}

VOID Insert (SOLUTIONLIST & list,
             CONST INTERVAL_VECTOR & vec, CONST INTERVAL & value)
{
  SOLUTIONLIST_ELEMENT iv(value, vec);
  list <<= iv;
}

VOID Append_First (SOLUTIONLIST & list,
                   CONST INTERVAL_VECTOR & vec, CONST INTERVAL & value)
{
  SOLUTIONLIST_ELEMENT iv(value, vec);
  list *= iv;
}

VOID Append (SOLUTIONLIST & dest, SOLUTIONLIST & source)
{
  while (!IsEmpty (source)) {
    dest += First (source);
    --source;
  }
}
