/************************************************************************
 *
 * Implementation of Several Useful Routines
 * -----------------------------------------
 *
 * Copyright (c) 1995, 1997 Olaf Knueppel
 *
 * $Id: VecUtils.C,v 1.1 1997/09/23 11:52:11 knueppel Exp $
 *
 ************************************************************************/

static char rcs_id[] = "$Id: VecUtils.C,v 1.1 1997/09/23 11:52:11 knueppel Exp $";

#include <GlobalOpt/VecUtils.h>

INT MaxDiam (CONST INTERVAL_VECTOR & v)
// Returns direction with maximal diameter, i.e.
// i with Diam (v(i)) >= Diam (v(j)) for all j
{
  INT i = 1;
  REAL di = Diam (v(i));
  REAL dj;

  for (INT j = 2; j <= Dimension(v); j++)
    if ((dj = Diam (v(j))) > di) { i = j; di = dj; }
  return i;
}

VOID SortDiam (CONST INTERVAL_VECTOR & x, INTEGER_VECTOR & maxdir)
// Generates an index vector "maxdir" which contains in decreasing diameters
// the indices of the components of x.
{
  INT i, j;
  INT dim = Dimension (x);
  INT tmp;
  REAL md, curmax;
  VECTOR maxdiam(dim);

  Resize(maxdir, dim);
  for (i = 1; i <= dim; i++) {
    maxdir(i) = i;
    maxdiam(i) = Diam (x(i));
  }
  for (i = 1; i < dim; i++) {
    curmax = maxdiam (maxdir(i));
    for (j = i + 1; j <= dim; j++) {
      md = maxdiam (maxdir(j));
      if (md > curmax)
	{
	  curmax = md;
	  tmp = maxdir(i);
	  maxdir(i) = maxdir(j);
	  maxdir(j) = tmp;
	}
    }
  }
}
