/************************************************************************
 *
 * Test Program for Non-Linear Function Type demonstration
 * -------------------------------------------------------
 *
 * Copyright (c) 1997 Olaf Knueppel
 *
 * $Id: TestNLF.C,v 1.3 1999/07/28 08:54:48 knueppel Exp $
 *
 ************************************************************************/

#include <NLF/NLF.h>
#include <Functions.h>
#include <Constants.h>
#include <Utilities.h>

/*
 * Definition of function
 */

static INTERVAL_AUTODIFF fFunctionTest (CONST INTERVAL_AUTODIFF & x, int n)
{
  switch (n) {
    case 1:
      return 1.0 - 2.0 * x(2) + 0.05 * Sin (4.0 * Constant::Pi * x(2)) - x(1);
    case 2:
      return x(2) - 0.5 * Sin (2.0 * Constant::Pi * x(1));
  }
  return 0.0 * x(1); /* dummy value */
}

FUNCTION FunctionTest (2, 2, fFunctionTest);

/*
 * Application
 */

VECTOR Newton (FUNCTION & f, VECTOR & x)
{
  VECTOR xAlt;

  do {
    xAlt = x;
    x = x - Inverse (Jacobian (f, x)) * VectorFunction (f, x);
  } while (Norm (x - xAlt) > 1e-6);
  return x;
}

/*
 * Call of application with function defined above
 */

main ()
{
  VECTOR Start(2);
  Start(1) = 0.0;
  Start(2) = 0.0;
  cout << "Approximation = "
       << Newton (FunctionTest, Start)
       << endl;
  return 0;
}
